<?php declare(strict_types=1);
/**
 * Twig Extension Class: for handling the environment of Twig
 */

namespace Infomaker\Everyware\Twig;

use Infomaker\Everyware\Support\NewRelicLog;
use Twig\Environment;
use Twig\Error\LoaderError;
use Twig\Error\RuntimeError;
use Twig\Error\SyntaxError;
use Twig\Loader\ChainLoader;
use Twig\Loader\LoaderInterface;
use Twig\Extension\DebugExtension;

/**
 * Class EveryTwig
 * @link    http://infomaker.se
 * @package Infomaker\Everyware\Twig
 */
class EveryTwig
{

    /**
     * Contains an instance of the parent for calls to certain functions
     *
     * @var Environment
     */
    private $twigEnvironment;

    /**
     * Contains an instance of self
     *
     * @var EveryTwig
     */
    /*  private static $instance; */

    /**
     * Contains a list of registered loaders
     *
     * @var array
     */
    private $loaders = [];

    /**
     * Contains whether to use cache
     *
     * @var bool
     */
    private $cache = false;

    /**
     * Time in seconds to cache the output
     *
     * @var int
     */
    private $cacheTime = 60;

    /**
     * Add suffix so we don't have to delete it in deploy every time
     *
     * @var string
     */
    private $cacheKey = 'everyware_twig_compiled_cache';

    /**
     * @var EveryTwig
     */
    private static $instance;

    private function __construct()
    {
        if (defined('GIT_COMMIT')) {
            $this->cacheKey .= '_' . GIT_COMMIT;
        }
    }

    /**
     * Create new instance if none exist
     *
     * @return EveryTwig
     */
    public static function getInstance(): EveryTwig
    {
        if (static::$instance === null) {
            static::$instance = new static;
        }

        return static::$instance;
    }

    /**
     *Initializes the EveryTwig-wrapper and Twig environment
     *
     * @param LoaderInterface $loader
     * @param array               $functions
     * @param array               $filters
     * @param array               $options
     *
     * @return Environment
     */
    public function init(
        LoaderInterface $loader,
        array $functions = [],
        array $filters = [],
        array $options = ['debug' => false, 'cache' => false]
    ): Environment {

        // Handle options
        $this->handleOptions($options);

        $this->addInitialLoader($loader);

        $chain = new ChainLoader($this->loaders);
        $this->twigEnvironment = new Environment($chain, $options);

        $this->twigEnvironment->addExtension(new DebugExtension());

        foreach ($functions as $function) {
            $this->twigEnvironment->addFunction($function);
        }

        foreach ($filters as $filter) {
            $this->twigEnvironment->addFilter($filter);
        }

        return $this->twigEnvironment;
    }

    /**
     * Handles the specific cache options and trims them away from the options array.
     *
     * @param array $options
     *
     * @return void
     */
    private function handleOptions(array &$options): void
    {
        if (isset($options['cachetime']) && (int)$options['cachetime'] > 0) {
            $this->cacheTime = (int)$options['cachetime'];
            unset($options['cachetime']);
        }

        if (isset($options['cache'])) {
            $this->cache = $options['cache'];
            // Add custom cache system to Twig
            $options['cache'] = $this->cache ? ObjectCacheSystem::setup($this->cacheKey, $this->cacheTime) : false;
        }
    }

    /**
     * Helper function to add a Twig_LoaderInterface loader on init
     *
     * @param LoaderInterface $loader
     *
     * @return void
     */
    private function addInitialLoader(LoaderInterface $loader): void
    {
        $this->loaders[] = $loader;
    }

    /**
     * Render function of EveryTwig, loads a specific template and works with cached compiled templates.
     *
     * @param string $template
     * @param array  $context
     *
     * @return void
     */
    public function render(string $template, array $context = null): void
    {
        try {
            $this->twigEnvironment->display($template, $context ?? []);
        } catch (LoaderError $e) {
            NewRelicLog::error("Cannot find twig-template: {$template}. ", $e);
        } catch (RuntimeError $e) {
            NewRelicLog::error("Failed to render twig-template: {$template}. ", $e);
        } catch (SyntaxError $e) {
            NewRelicLog::error("Failed to compile twig-template: {$template}. ", $e);
        }
    }

    /**
     * Return the view as rendered HTML by passed template
     *
     * @param string $template
     * @param array  $context
     *
     * @return string
     */
    public function view(string $template, array $context = null): string
    {
        try {
            return $this->twigEnvironment->render($template, $context ?? []);
        } catch (LoaderError $e) {
            NewRelicLog::error("Cannot find twig-template: {$template}. ", $e);
        } catch (RuntimeError $e) {
            NewRelicLog::error("Failed to render twig-template: {$template}. ", $e);
        } catch (SyntaxError $e) {
            NewRelicLog::error("Failed to compile twig-template: {$template}. ", $e);
        }

        return '';
    }
}
