<?php declare(strict_types=1);

namespace Infomaker\Everyware\Twig;

use Exception;
use Infomaker\Everyware\Support\NewRelicLog;
use Infomaker\Everyware\Twig\Extension\ProjectExtension;
use Twig\Environment;
use Twig\Error\LoaderError;
use Twig\Error\RuntimeError;
use Twig\Error\SyntaxError;
use Twig\Extension\DebugExtension;
use Twig\Extension\ExtensionInterface;
use Twig\Loader\LoaderInterface;
use Twig\NodeVisitor\NodeVisitorInterface;
use Twig\RuntimeLoader\RuntimeLoaderInterface;
use Twig\TokenParser\TokenParserInterface;
use Twig\TwigFilter;
use Twig\TwigFunction;
use Twig\TwigTest;
use Twig\Util\DeprecationCollector;

class ViewSetup
{
    /**
     * Add suffix so we don't have to delete it in deploy every time
     *
     * @var string
     */
    private $cacheKey = 'everyware_twig_compiled_cache';

    /**
     * Cache for 24h in seconds
     * @var int
     */
    protected $cacheTime = 86400;

    /** @var bool */
    protected $cache;

    /** @var bool */
    protected $debug;

    /**
     * Contains the FilesystemLoader aka all template-paths available in the templates.
     * @var LoaderInterface
     */
    protected $loader;

    private static $instance;

    /**
     * @var Environment
     */
    private $twigEnvironment;

    /**
     * @var TwigExtendable
     */
    private $localStorage;

    /**
     * Contains all registered template directories
     * @var array
     */
    private $registeredDirs = [];

    private function __construct(LoaderInterface $loader, TwigExtendable $localStorage)
    {
        $this->loader = $loader;
        $this->localStorage = $localStorage;
        $this->cache = ! is_dev() ? ObjectCacheSystem::setup($this->cacheKey, $this->cacheTime) : false;
        $this->debug = is_dev();

        if (defined('GIT_COMMIT')) {
            $this->cacheKey .= '_' . GIT_COMMIT;
        }
    }

    /**
     * Create new instance if none exist
     *
     * @return ViewSetup
     */
    public static function getInstance(): ViewSetup
    {
        if (static::$instance === null) {
            $instance = new static(new \Twig\Loader\FilesystemLoader(), new TwigExtenderStorage());

            // Set up the default Extensions
            $instance->addExtension(new ProjectExtension());

            if (is_dev()) {
                $instance->addExtension(new DebugExtension());
            }

            static::$instance = $instance;
        }

        return static::$instance;
    }

    public function setLoader(LoaderInterface $loader): void
    {
        @trigger_error(sprintf(__CLASS__ . '::setLoader is deprecated as of infomaker/everyware-twig v0.7.0.'),
            E_USER_DEPRECATED);
    }

    public function setEveryTwig(EveryTwig $everyTwig): void
    {
        @trigger_error(sprintf(__CLASS__ . '::setEveryTwig is deprecated as of infomaker/everyware-twig v0.7.0.'),
            E_USER_DEPRECATED);
    }

    private function getLoader(): LoaderInterface
    {
        @trigger_error(sprintf(__CLASS__ . '::getLoader is deprecated as of infomaker/everyware-twig v0.7.0.'),
            E_USER_DEPRECATED);

        return $this->loader;
    }

    public function getEveryTwig(): EveryTwig
    {
        @trigger_error(sprintf(__CLASS__ . '::getEveryTwig is deprecated as of infomaker/everyware-twig v0.7.0.'),
            E_USER_DEPRECATED);

        return EveryTwig::getInstance();
    }

    public function init(array $options = []): ViewSetup
    {
        if ($this->environmentInitialized()) {
            return $this;
        }

        $twig = new Environment($this->loader, array_replace([
            'debug' => $this->debug,
            'cache' => $this->cache
        ], $options));

        if ($this->localStorage instanceof TwigEnvironmentExtender) {
            $twig = $this->localStorage->extendEnvironment($twig);
        }

        $deprecationHandler = new DeprecationHandler($twig);

        if ($this->debug) {
            $deprecationHandler->searchAndLog($this->loader);
        } else {
            $deprecationHandler->listen();
        }

        $this->twigEnvironment = $twig;

        return $this;
    }

    /**
     * @param string     $namespace
     * @param string     $path
     * @param array|null $exclude
     *
     * @return ViewSetup
     */
    public function registerTwigFolder(string $namespace, string $path, array $exclude = null): self
    {
        // Store namespaces for registration
        if ( ! isset($this->registeredDirs[$namespace])) {
            $this->registeredDirs[$namespace] = [];
        }

        // Don't register paths twice under same namespace
        if (in_array($path, $this->registeredDirs[$namespace], true)) {
            return $this;
        }

        if (is_dir($path)) {
            try {
                if ($exclude !== null) {
                    @trigger_error(sprintf('Using "$exclude" in %s::registerTwigFolder is deprecated as of infomaker/everyware-twig v0.7.0.',
                        __CLASS__), E_USER_DEPRECATED);
                }

                $this->registeredDirs[$namespace][] = $path;

                $this->loader->addPath($path, $namespace);

            } catch (Exception $e) {
                NewRelicLog::error('Failed to register twig directory: ' . $path, $e);
            }
        }

        return $this;
    }

    /**
     * Render function of EveryTwig, loads a specific template and works with cached compiled templates.
     *
     * @param string $template
     * @param array  $context
     *
     * @return void
     */
    public function render(string $template, array $context = null): void
    {
        try {
            $this->twigEnvironment->display($template, $context ?? []);
        } catch (LoaderError $e) {
            NewRelicLog::error("Cannot find twig-template: {$template}. ", $e);
        } catch (RuntimeError $e) {
            NewRelicLog::error("Failed to render twig-template: {$template}. ", $e);
        } catch (SyntaxError $e) {
            NewRelicLog::error("Failed to compile twig-template: {$template}. ", $e);
        }
    }

    /**
     * Return the view as rendered HTML by passed template
     *
     * @param string $template
     * @param array  $context
     *
     * @return string
     */
    public function view(string $template, array $context = null): string
    {
        try {
            return $this->twigEnvironment->render($template, $context ?? []);
        } catch (LoaderError $e) {
            NewRelicLog::error("Cannot find twig-template: {$template}. ", $e);
        } catch (RuntimeError $e) {
            NewRelicLog::error("Failed to render twig-template: {$template}. ", $e);
        } catch (SyntaxError $e) {
            NewRelicLog::error("Failed to compile twig-template: {$template}. ", $e);
        }

        return '';
    }

    private function environmentInitialized(): bool
    {
        return $this->twigEnvironment !== null;
    }

    private function getExtender(): TwigExtendable
    {
        // If environment has yet to be initialized we'll store it for when it is
        return ! $this->environmentInitialized() ? $this->localStorage : new TwigExtender($this->twigEnvironment);
    }

    public function addExtension(ExtensionInterface $extension): void
    {
        $this->getExtender()->addExtension($extension);
    }

    public function addFilter(string $name, $callable = null, array $options = []): ViewSetup
    {
        $this->getExtender()->addFilter(new TwigFilter($name, $callable, $options));

        return $this;
    }

    public function addFunction(string $name, $callable = null, array $options = []): ViewSetup
    {
        $this->getExtender()->addFunction(new TwigFunction($name, $callable, $options));

        return $this;
    }

    public function addGlobal(string $key, $value): ViewSetup
    {
        $this->getExtender()->addGlobal($key, $value);

        return $this;
    }

    public function addNodeVisitor(NodeVisitorInterface $visitor): void
    {
        $this->getExtender()->addNodeVisitor($visitor);
    }

    public function addRuntimeLoader(RuntimeLoaderInterface $loader): void
    {
        $this->getExtender()->addRuntimeLoader($loader);
    }

    public function addTest(string $name, $callable = null, array $options = []): ViewSetup
    {
        $this->getExtender()->addTest(new TwigTest($name, $callable, $options));

        return $this;
    }

    public function addTokenParser(TokenParserInterface $parser): void
    {
        $this->getExtender()->addTokenParser($parser);
    }

    public function addWpFunction(string $functionName, string $name = null): ViewSetup
    {
        $name = ( ! empty($name) && $name !== null) ? $name : $functionName;
        $this->addFunction($name, $functionName);

        return $this;
    }

    public function hasExtension(string $class): bool
    {
        return $this->getExtender()->hasExtension($class);
    }
}
